/*
 * Copyright (c) 1996, 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#ifdef VTK

#ifndef MAKEDEPEND
#include <vtkActorCollection.h>
#include <vtkCamera.h>
#include <vtkCleanPolyData.h>
#include <vtkDataSetCollection.h>
#include <vtkDataSetReader.h>
#include <vtkFloatPoints.h>
#include <vtkFloatTCoords.h>
#include <vtkLODActor.h>
#include <vtkLight.h>
#include <vtkLine.h>
#include <vtkMath.h>
#include <vtkPNMReader.h>
#include <vtkPointSet.h>
#include <vtkPolyData.h>
#include <vtkPolyLine.h>
#include <vtkPolyDataMapper.h>
#include <vtkPolyDataWriter.h>
#include <vtkRenderer.h>
#include <vtkRendererCollection.h>
#include <vtkStructuredGrid.h>
#include <vtkTexture.h>
#include <vtkVRMLExporter.h>
#include <vtkWarpScalar.h>
#include <vtkRenderWindow.h>
#endif

#ifdef HAVE_ACE
#include <ace/SOCK_Stream.h>
#include <ace/SOCK_Connector.h>
#include <ace/INET_Addr.h>
#endif

extern "C" {
#include <projects.h>
}

#include <Dem/usgsdem.h>
#include <DemVtk/demdataset.h>
#include <DemVtk/demprobefilter.h>
#include <DemVtk/demtriangle.h>
#include <Map/mapfclass.h>
#include <Map/mapfeature.h>
#include <Map/mapprim.h>
#include <MapUnidraw/mapcomps.h>
#include <DemUnidraw/demcmds.h>
#include <DemUnidraw/demeditor.h>
#include <DemUnidraw/demviewer.h>
#include <DemUnidraw/lodactor.h>
#include <TopoFace/fgeomobjs.h>
#include <Unidraw/catalog.h>
#include <Unidraw/unidraw.h>
#include <InterViews/cursor.h>
#include <InterViews/session.h>
#include <InterViews/style.h>
#include <InterViews/window.h>
#include <IVGlyph/bdvalue.h>
#include <IVGlyph/importchooser.h>
#include <IVGlyph/ofilechooser.h>
#include <IV-look/kit.h>
#include <OS/string.h>
#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>
#include <math.h>
#include <stdio.h>
#include <sys/stat.h>
#undef Table

boolean viewup = false;
vtkDataSetCollection* datasets = NULL;

RoadRibbonFilter::RoadRibbonFilter() : vtkRibbonFilter() {}

void RoadRibbonFilter::Execute()
{
  int i, j;
  vtkPoints *inPts;
  vtkNormals *inNormals;
  vtkPointData *pd, *outPD;
  vtkCellArray *inLines;
  int numNewPts = 0;
  vtkFloatPoints *newPts;
  vtkFloatNormals *newNormals;
  vtkCellArray *newStrips;
  int npts, *pts;
  float p[3], pNext[3];
  float *n;
  float s[3], sNext[3], sPrev[3], w[3], tempw[3];
  double BevelAngle;
  float theta;
  int deleteNormals=0, ptId;
  vtkPolyData *input=(vtkPolyData *)this->Input;
  vtkPolyData *output=(vtkPolyData *)this->Output;
  vtkScalars *inScalars=NULL;
  float sFactor=1.0, range[2];
  int ptOffset=0;
//
// Initialize
//
  vtkDebugMacro(<<"Creating ribbon");

  inPts=input->GetPoints();
  inLines = input->GetLines();
  if ( !inPts || 
       (numNewPts=inPts->GetNumberOfPoints()*2) < 1 ||
       !inLines || inLines->GetNumberOfCells() < 1 )
    {
    vtkErrorMacro(<< ": No input data!\n");
    return;
    }

  // copy scalars, vectors, tcoords. Normals may be computed here.
  pd = input->GetPointData();
  outPD = output->GetPointData();
  outPD->CopyNormalsOff();
  outPD->CopyAllocate(pd,numNewPts);

  output->GetPointData()->CopyAllocate(pd,numNewPts);

  if ( !(inNormals=pd->GetNormals()) )
    {
    vtkPolyLine lineNormalGenerator;
    deleteNormals = 1;
    inNormals = vtkFloatNormals::New();
    ((vtkFloatNormals *)inNormals)->Allocate(numNewPts);
    if ( !lineNormalGenerator.GenerateSlidingNormals(inPts,inLines,(vtkFloatNormals*)inNormals) )
      {
      vtkErrorMacro(<< "No normals for line!\n");
      inNormals->Delete();
      return;
      }
    }
//
// If varying width, get appropriate info.
//
  if ( this->VaryWidth && (inScalars=pd->GetScalars()) )
    {
    inScalars->GetRange(range);
    }

  newPts = vtkFloatPoints::New();
  newPts->Allocate(numNewPts);
  newNormals = vtkFloatNormals::New();
  newNormals->Allocate(numNewPts);
  newStrips = vtkCellArray::New();
  newStrips->Allocate(newStrips->EstimateSize(1,numNewPts));
//
//  Create pairs of points along the line that are later connected into a 
//  triangle strip.
//
  theta = this->Angle * vtkMath::DegreesToRadians();
  for (inLines->InitTraversal(); inLines->GetNextCell(npts,pts); )
    {
//
// Use "averaged" segment to create beveled effect. Watch out for first and 
// last points.
//
    for (j=0; j < npts; j++)
      {

      if ( j == 0 ) //first point
        {
        inPts->GetPoint(pts[0],p);
        inPts->GetPoint(pts[1],pNext);
        for (i=0; i<3; i++) 
          {
          sNext[i] = pNext[i] - p[i];
          sPrev[i] = sNext[i];
          }
        }

      else if ( j == (npts-1) ) //last point
        {
        for (i=0; i<3; i++) 
          {
          sPrev[i] = sNext[i];
          p[i] = pNext[i];
          }
        }

      else
        {
        for (i=0; i<3; i++) p[i] = pNext[i];
        inPts->GetPoint(pts[j+1],pNext);
        for (i=0; i<3; i++)
          {
          sPrev[i] = sNext[i];
          sNext[i] = pNext[i] - p[i];
          }
        }

      n = inNormals->GetNormal(pts[j]);

      if ( vtkMath::Normalize(sNext) == 0.0 )
        {
        vtkErrorMacro(<<"Coincident points!");
        return;
        }

      for (i=0; i<3; i++) s[i] = (sPrev[i] + sNext[i]) / 2.0; //average vector
      vtkMath::Normalize(s);
      
      if ( (BevelAngle = vtkMath::Dot(sNext,sPrev)) > 1.0 ) BevelAngle = 1.0;
      if ( BevelAngle < -1.0 ) BevelAngle = -1.0;
      BevelAngle = acos((double)BevelAngle) / 2.0; //(0->90 degrees)
      if ( (BevelAngle = cos(BevelAngle)) == 0.0 ) BevelAngle = 1.0;

      BevelAngle = this->Width / BevelAngle;

      /* if the normal is bad, just use the last one */
      vtkMath::Cross(s,n,w);
      if ( vtkMath::Normalize(w) == 0.0)
        {
	  //vtkErrorMacro(<<"Bad normal!");
        w[0] = tempw[0];
        w[1] = tempw[1];
        w[2] = tempw[2];
        }
      tempw[0] = w[0];
      tempw[1] = w[1];
      tempw[2] = w[2];

      /* this seems to work, but it probably ain't right */
      float ty = w[1];
      float tz = w[2];
      w[2] = 0 /* ty */;
      w[1] = tz;
      vtkMath::Normalize(w);

      if ( inScalars )
        sFactor = 1.0 + ((this->WidthFactor - 1.0) * 
                  (inScalars->GetScalar(j) - range[0]) / (range[1]-range[0]));

      for (i=0; i<3; i++) s[i] = p[i] + w[i] * BevelAngle * sFactor;
      ptId = newPts->InsertNextPoint(s);
      newNormals->InsertNormal(ptId,n);
      outPD->CopyData(pd,pts[j],ptId);

      for (i=0; i<3; i++) s[i] = p[i] - w[i] * BevelAngle * sFactor;
      ptId = newPts->InsertNextPoint(s);
      newNormals->InsertNormal(ptId,n);
      outPD->CopyData(pd,pts[j],ptId);
      }
//
// Generate the strip topology
//
    newStrips->InsertNextCell(npts*2);
    for (i=0; i < npts; i++) 
      {//order important for consistent normals
      newStrips->InsertCellPoint(ptOffset+2*i+1);
      newStrips->InsertCellPoint(ptOffset+2*i);
      }
    
    ptOffset += npts*2;
    } //for this line
//
// Update ourselves
//
  if ( deleteNormals ) inNormals->Delete();

  output->SetPoints(newPts);
  newPts->Delete();

  output->SetStrips(newStrips);
  newStrips->Delete();

  outPD->SetNormals(newNormals);
  newNormals->Delete();

  output->Squeeze();
}

/*****************************************************************************/

DemTriangle* cell(float px, float py) {
    float p[3];
    p[0] = px;
    p[1] = py;
    p[2] = 0.0;
    vtkDataSet* dataset;
    float tol2;
    int subId;
    float pcoords[3], weights[3];
    
    for (datasets->InitTraversal(); dataset = datasets->GetNextItem(); )
	if (dataset->FindPoint(p) >= 0) {
	    tol2 = dataset->GetLength();
	    tol2 = tol2*tol2 / 1000.0;
	    DemTriangle* cell = (DemTriangle*)dataset->FindAndGetCell
	      (p, NULL, 0, tol2, subId, pcoords, weights);
	    if (cell) {
	        return cell;
	    }
	}
    return nil;
}

float /*DemCameraCmd::*/elev(float px, float py) {
#if 0
    float p[3];
    p[0] = px;
    p[1] = py;
    p[2] = 0.0;
    vtkDataSet* dataset;
    for (datasets->InitTraversal(); dataset = datasets->GetNextItem(); )
	if (dataset->FindPoint(p) >= 0) {
	    vtkFloatPoints* point = vtkFloatPoints::New();
	    ((vtkFloatPoints*)point)->Allocate(1);
	    point->InsertNextPoint(p);
	    vtkStructuredGrid* grid = vtkStructuredGrid::New();
	    grid->SetDimensions(1, 1, 1);
	    grid->SetPoints(point);
	    point->Delete();
	    DemProbeFilter* pfilter = DemProbeFilter::New();
	    pfilter->SetInput(grid);
	    pfilter->SetSource(dataset);
	    pfilter->Update();
	    vtkScalars* sc = pfilter->GetOutput()->
		GetPointData()->GetScalars();
	    if (sc)
		return sc->GetScalar(0);
	    else {
		printf("Warning: no scalars!\n");
		return 0.0;
	    }
	}
    //    printf("Warning: no elevation found!\n");
    return 0.0;
#endif
    float val = 0.0;
    int* x;
    int* y;
    int n;
    float p[3];
    p[0] = px;
    p[1] = py;
    p[2] = 0.0;
    DemEditor::GetDataSetXY(px, py, x, y, n);
    for (int i = 0; i < n; i++) {
      DemDataSet* ds = (DemDataSet*)DemEditor::GetDataSet(x[i], y[i], 0);
      if (ds->FindPoint(p) >= 0) {
	vtkFloatPoints* point = vtkFloatPoints::New();
	((vtkFloatPoints*)point)->Allocate(1);
	point->InsertNextPoint(p);
	vtkStructuredGrid* grid = vtkStructuredGrid::New();
	grid->SetDimensions(1, 1, 1);
	grid->SetPoints(point);
	point->Delete();
	DemProbeFilter* pfilter = DemProbeFilter::New();
	pfilter->SetInput(grid);
	pfilter->SetSource(ds);
	pfilter->Update();
	vtkScalars* sc = pfilter->GetOutput()->
	  GetPointData()->GetScalars();
	if (sc)
	  val = max(val, sc->GetScalar(0));
      }
    }
    return val;
}

/*****************************************************************************/

RibbonFixFilter::RibbonFixFilter() : vtkPolyDataToPolyDataFilter() {}

void RibbonFixFilter::Execute()
{
  this->GetOutput()->CopyStructure(this->GetInput());
  this->GetOutput()->GetPointData()->PassData(this->GetInput()->GetPointData());
  vtkFloatPoints* pts = (vtkFloatPoints*)this->GetOutput()->GetPoints();
  int npts = pts->GetNumberOfPoints();
  float pt[3];
  for (int i = 0; i < npts; i++) {
    pts->GetPoint(i, pt);
    pt[2] = elev(pt[0], pt[1]) + 0.1;
    pts->SetPoint(i, pt);
  }
}

/*****************************************************************************/

CameraCmd::CameraCmd(float px, float py, float fx, float fy, boolean project)
:Command() {
    px_ = px;
    py_ = py;
    fx_ = fx;
    fy_ = fy;
    project_ = project;
}

void CameraCmd::Execute() {
    Editor* ed = GetEditor();

    GetEditor()->GetWindow()->cursor(hourglass);
    DemCameraUpdateCmd* vcucmd = new DemCameraUpdateCmd(nil, 
							px_, py_, fx_, fy_,
							project_);
    vcucmd->SetEditor(ed);
    vcucmd->Execute();
    GetEditor()->GetWindow()->cursor(arrow);
}
     
/*****************************************************************************/

WaterLandLookupTable::WaterLandLookupTable()
: vtkLookupTable(2, 10) {
  SetTableValue(0, 135.0/255.0, 206.0/255.0, 250.0/255.0);
  SetTableValue(1, 255.0/255.0, 218.0/255.0, 185.0/255.0);
}

unsigned char* WaterLandLookupTable::MapValue(float v) {
    if (v == 0.0)
	return this->Table.GetColor(0);
    else
	return this->Table.GetColor(1);
}

/*****************************************************************************/

DemCmd::DemCmd(ControlInfo* ci) : Command(ci) {}

Command* DemCmd::Copy() {
    Command* copy = new DemCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void DemCmd::get_camera (float& px, float& py, float& pz,
		 float& fx, float& fy, float& fz) {
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkCamera* camera = ren->GetActiveCamera();
    float* pos = camera->GetPosition();
    px = pos[0];
    py = pos[1];
    pz = pos[2];
    float* foc = camera->GetFocalPoint();
    fx = foc[0];
    fy = foc[1];
    fz = foc[2];
}

/*****************************************************************************/

DemCameraCmd::DemCameraCmd(ControlInfo* ci) : DemCmd(ci) {}

Command* DemCameraCmd::Copy() {
    Command* copy = new DemCameraCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void DemCameraCmd::set_camera (float px, float py, float pz, float fx, float fy, float fz) {
    GetEditor()->GetWindow()->cursor(hourglass);
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkCamera* camera = ren->GetActiveCamera();
    camera->SetPosition(px, py, pz);
    camera->SetFocalPoint(fx, fy, fz);
    camera->ComputeViewPlaneNormal();  // necessary for proper camera setup
    camera->SetViewUp(0, 0, 1);
    camera->OrthogonalizeViewUp();  // this too is necessary
    float range[2];
    camera->GetClippingRange(range);
    camera->SetClippingRange(1.0, /*range[1]*/5000.0);
    update_states();
    GetEditor()->GetViewer()->Draw();
}

void DemCameraCmd::yaw_camera (float ang) {
    GetEditor()->GetWindow()->cursor(hourglass);
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkCamera* camera = ren->GetActiveCamera();
    camera->Yaw(ang);
    update_states();
    GetEditor()->GetViewer()->Draw();
}

void DemCameraCmd::reset_camera () {
    GetEditor()->GetWindow()->cursor(hourglass);
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
#if 0
    const char* lod = unidraw->GetCatalog()->GetAttribute("lod");
    if (strcmp(lod,"false")==0)
      ren->ResetCamera();
#endif
    Attribute* lod = ((OverlayComp*)GetEditor()->GetComponent())
      ->GetAttributeList()->GetAttr("lod");
    if (!lod)
      ren->ResetCamera();
    update_states();
    GetEditor()->GetViewer()->Draw();
}

void DemCameraCmd::update_camera (/* vtkRenderer* ren, UsgsDem* dem, */
		    float px, float py, float fx, float fy, boolean project) {
    GetEditor()->GetWindow()->cursor(hourglass);
    DemEditor* ed = (DemEditor*)GetEditor();
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
//    UsgsDem* dem = GetEditor()->DEM();
    if (project) {
    char lon0[BUFSIZ];
    sprintf(lon0, "lon_0=%f", px * RAD_TO_DEG);  // let proj pick the utm zone
    char* parms[4];
    parms[0] = "proj=utm";
    parms[1] = "ellps=WGS72";
    parms[2] = lon0;
    parms[3] = "no_defs";
    PJ* proj = pj_init(4, parms);

    UV data;
    data.u = px;
    data.v = py;
//    data = pj_fwd(data, dem->projection());
    data = pj_fwd(data, proj);
    px = data.u;
    py = data.v;
    data.u = fx;
    data.v = fy;
//    data = pj_fwd(data, dem->projection());
    data = pj_fwd(data, proj);
    fx = data.u;
    fy = data.v;
    }
    //    printf("UTM px=%f, py=%f, fx=%f, fy=%f\n", px, py, fx, fy);
    vtkCamera* camera = ren->GetActiveCamera();
    float elevation = elev(px, py);
    //    printf("elevation at px,py: %f\n", elevation);
    camera->SetFocalPoint(fx, fy, elevation + ed->CameraHeight());
    camera->SetPosition(px, py, elevation + ed->CameraHeight());
    camera->ComputeViewPlaneNormal();  // necessary for proper camera setup
    camera->SetViewUp(0, 0, 1);
    camera->OrthogonalizeViewUp();  // this too is necessary
    viewup = true;
    float range[2];
    camera->GetClippingRange(range);
    camera->SetClippingRange(1.0, /*range[1]*/5000.0);

    update_states();
//    renWindow->Render();
    GetEditor()->GetViewer()->Draw();
}

void DemCameraCmd::update_states() {
    DemEditor* ed = (DemEditor*)GetEditor();
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkCamera* camera = ren->GetActiveCamera();

    if (camera) {
    float pos[3];
    float fpt[3];
    float range[2];
    camera->GetPosition(pos);
    camera->GetFocalPoint(fpt);
    camera->GetClippingRange(range);
    float viewup[3];
    camera->GetViewUp(viewup);
    float angle = camera->GetViewAngle();
    float dist = camera->GetDistance();
    //    printf("I am a camera:\n");
    //    printf("position: %f %f %f\n", pos[0], pos[1], pos[2]);
    //    printf("focal pt: %f %f %f\n", fpt[0], fpt[1], fpt[2]);
    //    printf("clipping range: front %f, back %f\n", range[0], range[1]);
    //    printf("view up: %f %f %f\n", viewup[0], viewup[1], viewup[2]);
    //    printf("view angle: %f\n", angle);
    //    printf("distance: %f\n\n", dist);
    ed->CameraX()->current_value(pos[0]);
    ed->CameraY()->current_value(pos[1]);
    ed->CameraZ()->current_value(pos[2]);
    ed->FocalPtX()->current_value(fpt[0]);
    ed->FocalPtY()->current_value(fpt[1]);
    ed->FocalPtZ()->current_value(fpt[2]);
    ed->ClipFront()->current_value(range[0]);
    ed->ClipBack()->current_value(range[1]);
    ed->ViewUpX()->current_value(viewup[0]);
    ed->ViewUpY()->current_value(viewup[1]);
    ed->ViewUpZ()->current_value(viewup[2]);
    ed->ViewAngle()->current_value(angle);
    ed->Distance()->current_value(dist);
    }
}

/*****************************************************************************/

DemLoadDemCmd::DemLoadDemCmd(ControlInfo* info)
: DemCameraCmd(info) {
    chooser_ = nil;
    fname_ = nil;
}

DemLoadDemCmd::DemLoadDemCmd(const char* dn)
: DemCameraCmd(nil) {
    chooser_ = nil;
    fname_ = strdup(dn);
}

DemLoadDemCmd::~DemLoadDemCmd() {}

Command* DemLoadDemCmd::Copy() {
    Command* copy = new DemLoadDemCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

boolean DemLoadDemCmd::Reversible() {
    return false;
}

void DemLoadDemCmd::Execute() {
    Style* style;
    boolean reset_caption = false;
    struct stat buf;
    if (fname_ == nil) {
    if (chooser_ == nil) {
	style = new Style(Session::instance()->style());
	chooser_ = new OpenFileChooser(".", WidgetKit::instance(), style);
	Resource::ref(chooser_);
	char buf[CHARBUFSIZE];
	const char* domain = unidraw->GetCatalog()->GetAttribute("domain");
	domain = (domain == nil) ? "component" : domain;
	sprintf(buf, "Select a vtk %s to open:", domain);
	style->attribute("caption", "");
	style->attribute("subcaption", buf);
    } else {
	style = chooser_->style();
    }
    while (chooser_->post_for(_editor->GetWindow())) {
	_editor->GetWindow()->cursor(hourglass);
	chooser_->twindow()->cursor(hourglass);
        const String* s = chooser_->selected();
	NullTerminatedString ns(*s);
	const char* name = ns.string();
	if (stat(name, &buf) != 0) {
	    style->attribute("caption", "");
	    style->attribute("caption", "Open failed: file not found");
	    reset_caption = true;
	    GetEditor()->GetWindow()->cursor(arrow);
	    chooser_->twindow()->cursor(arrow);
	    continue;
	}

	vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	    ->GetRenderWindow();
	vtkRendererCollection* rc = renWindow->GetRenderers();
	rc->InitTraversal();
	vtkRenderer* ren = rc->GetNextItem();
	WaterLandLookupTable* table = WaterLandLookupTable::New();

	vtkDataSetReader* reader = vtkDataSetReader::New();
	reader->SetFileName((char*)name);
	reader->Update();
	vtkPolyData* data = (vtkPolyData*)reader->GetOutput();
	if (!data) {
	    style->attribute("caption", "");
	    style->attribute("caption", "Read failed: polydata not found");
	    reset_caption = true;
	    GetEditor()->GetWindow()->cursor(arrow);
	    chooser_->twindow()->cursor(arrow);
	    continue;
        }
	DemDataSet* demds = DemDataSet::New();
	demds->SetPolyData(data);

	if (!datasets) {
	  datasets = vtkDataSetCollection::New();
	}
	//	datasets->AddItem(reader->GetOutput());
        datasets->AddItem(demds);

	boolean have_texture = false;
	const char* tfile = unidraw->GetCatalog()->GetAttribute("texture");
	if (strcmp(tfile,"")!=0)
	  have_texture = true;

	vtkTexture* texture;
	if (have_texture) {
	  // set up texture coordinates
	  int npts = demds->GetNumberOfPoints();
	  float bounds[6];
	  demds->GetBounds(bounds);
	  float xext = bounds[1] - bounds[0];
	  float yext = bounds[3] - bounds[2];
	  vtkFloatTCoords* tcoords = vtkFloatTCoords::New();
	  tcoords->Allocate(npts);
	  for (int i = 0; i < npts; i++) {
	    float point[3];
	    demds->GetPoint(i, point);
	    float tc[2];
	    tc[0] = ((point[0] - bounds[0]) / xext) * 20;
	    tc[1] = ((point[1] - bounds[2]) / yext) * 20;
	    tcoords->InsertTCoord(i, tc);
	  }
	  demds->GetPointData()->SetTCoords(tcoords);
	  tcoords->Delete();

	  vtkPNMReader* pnm = vtkPNMReader::New();
	  pnm->SetFileName((char*)tfile);
	  texture = vtkTexture::New();
	  texture->SetInput(pnm->GetOutput());
	  texture->InterpolateOn();
	  texture->RepeatOn();
	}

	vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
	mapper->SetLookupTable(table);
	mapper->SetInput(demds);

	LODActor* actor = LODActor::New();
	actor->SetMapper(mapper);
	if (have_texture)
	  actor->SetTexture(texture);

	//	actor->GetProperty()->SetWireframe();
	actor->GetProperty()->SetAmbient(0.15);
	actor->GetProperty()->SetDiffuse(0.4);
	actor->GetProperty()->SetAmbientColor(255.0/255.0, 218.0/255.0, 185.0/255.0);
	ren->AddActor(actor);
	if (!viewup) {
	    vtkCamera* cam = ren->GetActiveCamera();
// Just for initial view
//	    cam->Elevation(-90);
//	    viewup = true;
	}
	Attribute* lod = ((OverlayComp*)GetEditor()->GetComponent())
	  ->GetAttributeList()->GetAttr("lod");
	if (!lod)
	  //const char* lod = unidraw->GetCatalog()->GetAttribute("lod");
	  //if (strcmp(lod,"false")==0)
	  if (!viewup)
	    ren->ResetCamera();
	update_states();
	_editor->GetViewer()->Draw();
	break;
    }
    chooser_->unmap();
    if (reset_caption) {
	style->attribute("caption", "");
    }
    }
    else {
	_editor->GetWindow()->cursor(hourglass);
	vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	    ->GetRenderWindow();
	vtkRendererCollection* rc = renWindow->GetRenderers();
	rc->InitTraversal();
	vtkRenderer* ren = rc->GetNextItem();
	WaterLandLookupTable* table = WaterLandLookupTable::New();

	char prepath[255];
	prepath[0] = '\0';
	Catalog* catalog = unidraw->GetCatalog();
	char* url = (char*)catalog->GetAttribute("url");
	if (url && strlen(url) > 0) {
	  if (strstr(url, "file:") == url) {
	    strcpy(prepath, (char*)url + 5);
	    char* lastslashpos = strrchr(prepath, '/');
	    if (lastslashpos)
	      *(lastslashpos+1) = '\0';
	  }
	}
	strcat(prepath, (char*)fname_);

	vtkDataSetReader* reader = vtkDataSetReader::New();
	reader->SetFileName(prepath);
	reader->Update();
	vtkPolyData* data = (vtkPolyData*)reader->GetOutput();
	if (!data) {
	  return;
        }
	DemDataSet* demds = DemDataSet::New();
	demds->SetPolyData(data);

	if (!datasets) {
	  datasets = vtkDataSetCollection::New();
	}
        datasets->AddItem(demds);

	boolean have_texture = false;
	const char* tfile = unidraw->GetCatalog()->GetAttribute("texture");
	if (strcmp(tfile,"")!=0)
	  have_texture = true;

	vtkTexture* texture;
	if (have_texture) {
	  // set up texture coordinates
	  int npts = demds->GetNumberOfPoints();
	  float bounds[6];
	  demds->GetBounds(bounds);
	  float xext = bounds[1] - bounds[0];
	  float yext = bounds[3] - bounds[2];
	  vtkFloatTCoords* tcoords = vtkFloatTCoords::New();
	  tcoords->Allocate(npts);
	  for (int i = 0; i < npts; i++) {
	    float point[3];
	    demds->GetPoint(i, point);
	    float tc[2];
	    tc[0] = ((point[0] - bounds[0]) / xext) * 20;
	    tc[1] = ((point[1] - bounds[2]) / yext) * 20;
	    tcoords->InsertTCoord(i, tc);
	  }
	  demds->GetPointData()->SetTCoords(tcoords);
	  tcoords->Delete();

	  vtkPNMReader* pnm = vtkPNMReader::New();
	  pnm->SetFileName((char*)tfile);
	  pnm->Update();
	  texture = vtkTexture::New();
	  texture->SetInput(pnm->GetOutput());
	  texture->InterpolateOn();
	  texture->RepeatOn();
	}

	vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
	mapper->SetLookupTable(table);
	mapper->SetInput(demds);

	LODActor* actor = LODActor::New();
	actor->SetMapper(mapper);
	if (have_texture)
	  actor->SetTexture(texture);

	actor->GetProperty()->SetAmbient(0.15);
	actor->GetProperty()->SetDiffuse(0.4);
	actor->GetProperty()->SetAmbientColor(255.0/255.0, 218.0/255.0, 185.0/255.0);
	ren->AddActor(actor);
	if (!viewup) {
	    vtkCamera* cam = ren->GetActiveCamera();
// Just for initial view
//	    cam->Elevation(-90);
//	    viewup = true;
	}
	Attribute* lod = ((OverlayComp*)GetEditor()->GetComponent())
	  ->GetAttributeList()->GetAttr("lod");
	if (!lod)
	//const char* lod = unidraw->GetCatalog()->GetAttribute("lod");
	//if (strcmp(lod,"false")==0)
	  if (!viewup)
	    ren->ResetCamera();
	update_states();
	//	_editor->GetViewer()->Draw();
    }
}

/*****************************************************************************/

DemLoadLODDemCmd::DemLoadLODDemCmd(ControlInfo* info)
: Command(info) {
}

DemLoadLODDemCmd::~DemLoadLODDemCmd() {}

Command* DemLoadLODDemCmd::Copy() {
    Command* copy = new DemLoadLODDemCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

boolean DemLoadLODDemCmd::Reversible() {
    return false;
}

void DemLoadLODDemCmd::Execute() {
  _editor->GetWindow()->cursor(hourglass);
  vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
    ->GetRenderWindow();
  vtkRendererCollection* rc = renWindow->GetRenderers();
  rc->InitTraversal();
  vtkRenderer* ren = rc->GetNextItem();

  vtkTexture* texture;
  boolean have_texture = false;
  const char* tfile = unidraw->GetCatalog()->GetAttribute("texture");
  if (strcmp(tfile,"")!=0) {
    have_texture = true;
    vtkPNMReader* pnm = vtkPNMReader::New();
    pnm->SetFileName((char*)tfile);
    pnm->Update();
    texture = vtkTexture::New();
    texture->SetInput(pnm->GetOutput());
    texture->InterpolateOn();
    texture->RepeatOn();
  }

  vtkDataSet* firstds = nil;
  char* basename = "/data/maps/dem/sfslod/SANFRANCISCOSOUTH,CABIGBASINDEM.";
  char fnbuf[128];
  for (int xi = 0; xi < 4; xi++) {
    for (int yi = 0; yi < 4; yi++) {
      LODActor* actor = LODActor::New();
      if (have_texture)
	actor->SetTexture(texture);

      for (int lod = 0; lod < 6; lod++) {
	memset(fnbuf, 0, 128);
	strcpy(fnbuf, basename);
	fnbuf[strlen(fnbuf)] = '0' + xi;
	fnbuf[strlen(fnbuf)] = '.';
	fnbuf[strlen(fnbuf)] = '0' + yi;
	fnbuf[strlen(fnbuf)] = '.';
	fnbuf[strlen(fnbuf)] = '0' + lod;
	strcat(fnbuf, ".vtk");
	
	vtkDataSetReader* reader = vtkDataSetReader::New();
	reader->SetFileName(fnbuf);
	reader->Update();
	vtkPolyData* data = (vtkPolyData*)reader->GetOutput();
	if (!data) {
	  continue;
	}
	DemDataSet* demds = DemDataSet::New();
	demds->SetPolyData(data);

	if (!datasets) {
	  datasets = vtkDataSetCollection::New();
	}
	if (lod == 0)
	  datasets->AddItem(demds);
	if (!firstds)
	  firstds = demds;

	if (have_texture) {
	  // set up texture coordinates
	  int npts = demds->GetNumberOfPoints();
	  float bounds[6];
	  demds->GetBounds(bounds);
	  float xext = bounds[1] - bounds[0];
	  float yext = bounds[3] - bounds[2];
	  vtkFloatTCoords* tcoords = vtkFloatTCoords::New();
	  tcoords->Allocate(npts);
	  for (int i = 0; i < npts; i++) {
	    float point[3];
	    demds->GetPoint(i, point);
	    float tc[2];
	    tc[0] = ((point[0] - bounds[0]) / xext) * 20;
	    tc[1] = ((point[1] - bounds[2]) / yext) * 20;
	    tcoords->InsertTCoord(i, tc);
	  }
	  demds->GetPointData()->SetTCoords(tcoords);
	  tcoords->Delete();
	}

	vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
  WaterLandLookupTable* table = WaterLandLookupTable::New();
	mapper->SetLookupTable(table);
	mapper->SetInput(demds);

	actor->SetLODMapper(lod, mapper);
      }
      actor->GetProperty()->SetAmbient(0.15);
      actor->GetProperty()->SetDiffuse(0.4);
      actor->GetProperty()->SetAmbientColor(255.0/255.0, 218.0/255.0, 185.0/255.0);

      ren->AddActor(actor);
    }
  }
#if 0
  if (!viewup) {
    vtkCamera* cam = ren->GetActiveCamera();
    // Just for initial view
    cam->Elevation(-90);
    viewup = true;
  }
  _editor->GetViewer()->Draw();
#endif

  float bounds[6];
  firstds->GetBounds(bounds);
  float px = bounds[0] + (bounds[1] - bounds[0])/2.0;
  float py = bounds[2] + (bounds[3] - bounds[2])/2.0;
  float fx = bounds[0] + (bounds[1] - bounds[0])/2.0;
  float fy = bounds[3];
  DemCameraUpdateCmd* camupd = new DemCameraUpdateCmd(nil, px, py, fx, fy, false);
  camupd->SetEditor(GetEditor());
  camupd->Execute();
}

/*****************************************************************************/

DemImportFCCmd::DemImportFCCmd(ControlInfo* info, ImportChooser* ch)
  : MapImportCmd(info, ch) {
}

void DemImportFCCmd::Execute() {
    MapFClassComp* fccomp = 
        inptr_ ?  (MapFClassComp*)Import(*inptr_) : (MapFClassComp*)PostDialog();

    if (fccomp) {
        MapFeatureClass* fc = fccomp->GetFeatureClass();
	int nf = fc->nfeats();
	vtkPolyData* fcdataset = vtkPolyData::New();
	vtkFloatPoints* fcpoints = vtkFloatPoints::New();
	int pti = 0;
	vtkCellArray* fccells = vtkCellArray::New();
	for (int i = 0; i < nf; i++) {

	    MapFeature* f = fc->feature(i);
	    switch(f->type()) {
	    case MapFeature::PointType:
	      break;
	    case MapFeature::LineType: 
	      {
		MapEdge* edge = f->edge_primitive();
		int npts = edge->npts();
		const float* x = edge->xpoints();
		const float* y = edge->ypoints();
		vtkIdList* ids = vtkIdList::New();
		float last_elev;
//		cerr << "multiline(";
		for (int j = 0; j < npts; j++) {
		  float point[3];
		  point[0] = x[j];
		  point[1] = y[j];
		  point[2] = elev(x[j], y[j]) + .1;
		  
		  if (j>0) {
//		    cerr << ",";
		    double xlen = x[j] - x[j-1];    
		    double ylen = y[j] - y[j-1];    
		    double zlen = point[2] - last_elev;
		    const double rate = 100.0; // 50.0; // 10.0;
		    double distance = sqrt(xlen*xlen + ylen*ylen + zlen*zlen);
		    double ratio = rate/distance;
		    double dx = xlen * ratio;
		    double dy = ylen * ratio;
		    int times = 1;
		    double prevx = x[j-1];
		    double prevy = y[j-1];
		    double lastx = prevx;
		    double lasty = prevy;
		    for (double k=rate;k<distance;k+= rate) {
		      double curx = dx*times + prevx;
		      double cury = dy*times + prevy;

		      // if line crosses cell bdry then sample at the intrsct
		      FLineObj rline(lastx, lasty, curx, cury);
		      DemTriangle* t1 = cell(lastx, lasty);
		      if (t1) {
			// check cell edges
			float tx0 = t1->Points.GetPoint(0)[0];
			float ty0 = t1->Points.GetPoint(0)[1];
			float tx1 = t1->Points.GetPoint(1)[0];
			float ty1 = t1->Points.GetPoint(1)[1];
			float tx2 = t1->Points.GetPoint(2)[0];
			float ty2 = t1->Points.GetPoint(2)[1];
			FLineObj tline1(tx0, ty0, tx1, ty1);
			FLineObj tline2(tx1, ty1, tx2, ty2);
			FLineObj tline3(tx2, ty2, tx0, ty0);
			boolean inter = false;
			float intx, inty;
			if (rline.Intersects(tline1)) {
			  inter = rline.EquationIntersects(tline1, intx, inty);
			}
			else if (rline.Intersects(tline2)) {
			  inter = rline.EquationIntersects(tline2, intx, inty);
			}
			else if (rline.Intersects(tline3)) {
			  inter = rline.EquationIntersects(tline3, intx, inty);
			}
			if (inter && !isnan(intx)) {
			  //printf("adding extra point 1\n");
			  float ipoint[3];
			  ipoint[0] = intx;
			  ipoint[1] = inty;
			  ipoint[2] = elev(intx, inty) + .1;
			  fcpoints->InsertPoint(pti, ipoint);
			  ids->InsertNextId(pti++);
			}

			// check neighboring cells?
		      }
		      DemTriangle* t2 = cell(curx, cury);
		      if (t2) {
			// check cell edges
			float tx0 = t2->Points.GetPoint(0)[0];
			float ty0 = t2->Points.GetPoint(0)[1];
			float tx1 = t2->Points.GetPoint(1)[0];
			float ty1 = t2->Points.GetPoint(1)[1];
			float tx2 = t2->Points.GetPoint(2)[0];
			float ty2 = t2->Points.GetPoint(2)[1];
			FLineObj tline1(tx0, ty0, tx1, ty1);
			FLineObj tline2(tx1, ty1, tx2, ty2);
			FLineObj tline3(tx2, ty2, tx0, ty0);
			boolean inter = false;
			float intx, inty;
			if (rline.Intersects(tline1)) {
			  inter = rline.EquationIntersects(tline1, intx, inty);
			}
			else if (rline.Intersects(tline2)) {
			  inter = rline.EquationIntersects(tline2, intx, inty);
			}
			else if (rline.Intersects(tline3)) {
			  inter = rline.EquationIntersects(tline3, intx, inty);
			}
			float lastpt[3];
			fcpoints->GetPoint(pti-1, lastpt);
			if (inter && !isnan(intx) &&
			    intx != lastpt[0] && inty != lastpt[1]) {
			  //printf("adding extra point 2\n");
			  float ipoint[3];
			  ipoint[0] = intx;
			  ipoint[1] = inty;
			  ipoint[2] = elev(intx, inty) + .1;
			  fcpoints->InsertPoint(pti, ipoint);
			  ids->InsertNextId(pti++);
			}

			// check neighboring cells?
		      }

		      // sample next point
		      float npoint[3];
		      npoint[0] = curx;
		      npoint[1] = cury;
		      npoint[2] = elev(curx, cury) + .1;
		      fcpoints->InsertPoint(pti, npoint);
		      ids->InsertNextId(pti++);
//		      cerr << (int) npoint[0] << "," << (int) npoint[1] << ",";
		      times++;

		      // save last pt
		      lastx = curx;
		      lasty = cury;
		    }
		  } 
		  
		  fcpoints->InsertPoint(pti, point);
		  ids->InsertNextId(pti++);
		  last_elev = point[2];
//		  cerr << (int) point[0] << "," << (int) point[1];
		}
		vtkLine* line = vtkLine::New();
		line->PointIds = *ids;
		ids->Delete();
		fccells->InsertNextCell(line);
//		cerr << ")\n";
	      }
	      break;
	    case MapFeature::AreaType:
	      break;
	    case MapFeature::TextType:
	    case MapFeature::JunctionType:
	    case MapFeature::RasterType:
	    case MapFeature::DemType:
	      break;
	    }
	}
	fcdataset->SetPoints(fcpoints);
	fcpoints->Delete();
	fcdataset->SetLines(fccells);
	fccells->Delete();
	vtkCleanPolyData* clean = vtkCleanPolyData::New();
	clean->SetTolerance(0.0);
	clean->SetInput(fcdataset);

#if 1
	RoadRibbonFilter* rib = RoadRibbonFilter::New();
	rib->VaryWidthOff();
	rib->SetWidth(10.0);
	rib->SetWidthFactor(1.0);
	rib->SetAngle(90.0);
#else
	vtkTubeFilter* rib = vtkTubeFilter::New();
	rib->SetVaryRadius(VTK_VARY_RADIUS_OFF);
	rib->SetRadius(1.0);
	rib->SetRadiusFactor(1.0);
	rib->SetNumberOfSides(8);
#endif
	rib->SetInput(clean->GetOutput());
	RibbonFixFilter* ribfix = RibbonFixFilter::New();
	ribfix->SetInput(rib->GetOutput());

	vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
	//mapper->SetInput(fcdataset);
	mapper->SetInput(ribfix->GetOutput());
	mapper->ScalarVisibilityOff();
	// lookup table?
	LODActor* actor = LODActor::New();
	actor->SetMapper(mapper);
	vtkProperty* prop = vtkProperty::New();
	prop->SetInterpolationToFlat();
	prop->SetColor(1.0, 0.3, 0.3);
	prop->SetRepresentationToSurface();
	prop->SetAmbient(0.5);
	prop->SetDiffuse(0.5);
	actor->SetProperty(prop);

	vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	  ->GetRenderWindow();
	vtkRendererCollection* rc = renWindow->GetRenderers();
	rc->InitTraversal();
	vtkRenderer* ren = rc->GetNextItem();
	ren->AddActor(actor);
	GetEditor()->GetViewer()->Draw();

	vtkPolyDataWriter* fcwriter = vtkPolyDataWriter::New();
	fcwriter->SetInput(ribfix->GetOutput()/*fcdataset*/);
	//fcwriter->SetFilename("mapfc.vtk");
	const String* str = chooser_->selected();
	NullTerminatedString ns(*str);
	char buf[256];
	strcpy(buf, ns.string());
	strcat(buf, ".vtk");
	fcwriter->SetFileName(buf);
	fcwriter->SetFileType(VTK_ASCII);
	fcwriter->Write();

	fcwriter->Delete();
	//ribfix->Delete();
	//rib->Delete();
	//clean->Delete();
	//fcdataset->Delete();
    }
}

/*****************************************************************************/

DemLoadFCCmd::DemLoadFCCmd(ControlInfo* ci)
  : Command(ci) {
    fname_ = nil;
}

DemLoadFCCmd::DemLoadFCCmd(const char* dn)
  : Command() {
    fname_ = strdup(dn);
}

void DemLoadFCCmd::Execute() {
    Style* style;
    boolean reset_caption = false;
    struct stat buf;
    if (fname_ == nil) {
    if (chooser_ == nil) {
	style = new Style(Session::instance()->style());
	chooser_ = new OpenFileChooser(".", WidgetKit::instance(), style);
	Resource::ref(chooser_);
	char buf[CHARBUFSIZE];
	const char* domain = nil;//unidraw->GetCatalog()->GetAttribute("domain");
	domain = (domain == nil) ? "feature class" : domain;
	sprintf(buf, "Select a vtk %s to open:", domain);
	style->attribute("caption", "");
	style->attribute("subcaption", buf);
    } else {
	style = chooser_->style();
    }
    while (chooser_->post_for(_editor->GetWindow())) {
	_editor->GetWindow()->cursor(hourglass);
	chooser_->twindow()->cursor(hourglass);
        const String* s = chooser_->selected();
	NullTerminatedString ns(*s);
	const char* name = ns.string();
	if (stat(name, &buf) != 0) {
	    style->attribute("caption", "");
	    style->attribute("caption", "Open failed: file not found");
	    reset_caption = true;
	    GetEditor()->GetWindow()->cursor(arrow);
	    chooser_->twindow()->cursor(arrow);
	    continue;
	}

	vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	    ->GetRenderWindow();
	vtkRendererCollection* rc = renWindow->GetRenderers();
	rc->InitTraversal();
	vtkRenderer* ren = rc->GetNextItem();

	vtkDataSetReader* reader = vtkDataSetReader::New();
	reader->SetFileName((char*)name);
	reader->Update();
	vtkPolyData* data = (vtkPolyData*)reader->GetOutput();
	if (!data) {
	    style->attribute("caption", "");
	    style->attribute("caption", "Read failed: polydata not found");
	    reset_caption = true;
	    GetEditor()->GetWindow()->cursor(arrow);
	    chooser_->twindow()->cursor(arrow);
	    continue;
        }

	if (!datasets) {
	  datasets = vtkDataSetCollection::New();
	}
	vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
	mapper->SetInput(data);

	LODActor* actor = LODActor::New();
	actor->SetMapper(mapper);
	vtkProperty* prop = vtkProperty::New();
	prop->SetInterpolationToFlat();
	prop->SetColor(1.0, 0.3, 0.3);
	prop->SetRepresentationToSurface();
	prop->SetAmbient(0.5);
	prop->SetDiffuse(0.5);
	actor->SetProperty(prop);
	ren->AddActor(actor);
	_editor->GetViewer()->Draw();
	break;
    }
    chooser_->unmap();
    if (reset_caption) {
	style->attribute("caption", "");
    }
    }
    else {
	_editor->GetWindow()->cursor(hourglass);
	vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	    ->GetRenderWindow();
	vtkRendererCollection* rc = renWindow->GetRenderers();
	rc->InitTraversal();
	vtkRenderer* ren = rc->GetNextItem();

	char prepath[255];
	prepath[0] = '\0';
	Catalog* catalog = unidraw->GetCatalog();
	char* url = (char*)catalog->GetAttribute("url");
	if (url && strlen(url) > 0) {
	  if (strstr(url, "file:") == url) {
	    strcpy(prepath, (char*)url + 5);
	    char* lastslashpos = strrchr(prepath, '/');
	    if (lastslashpos)
	      *(lastslashpos+1) = '\0';
	  }
	}
	strcat(prepath, (char*)fname_);

	vtkDataSetReader* reader = vtkDataSetReader::New();
	reader->SetFileName(prepath);
	reader->Update();
	vtkPolyData* data = (vtkPolyData*)reader->GetOutput();
	if (!data) {
#if 0
	    style->attribute("caption", "");
	    style->attribute("caption", "Read failed: polydata not found");
	    reset_caption = true;
	    GetEditor()->GetWindow()->cursor(arrow);
	    chooser_->twindow()->cursor(arrow);
	    continue;
#endif
	    return;
        }

	if (!datasets) {
	  datasets = vtkDataSetCollection::New();
	}
	vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
	mapper->SetInput(data);

	LODActor* actor = LODActor::New();
	actor->SetMapper(mapper);
	vtkProperty* prop = vtkProperty::New();
	prop->SetInterpolationToFlat();
	prop->SetColor(1.0, 0.3, 0.3);
	prop->SetRepresentationToSurface();
	prop->SetAmbient(0.5);
	prop->SetDiffuse(0.5);
	actor->SetProperty(prop);
	ren->AddActor(actor);
	_editor->GetViewer()->Draw();
    }
}

/*****************************************************************************/

DemExportVRMLCmd::DemExportVRMLCmd(ControlInfo* info)
: Command(info) {
    chooser_ = nil;
}

DemExportVRMLCmd::~DemExportVRMLCmd() {}

Command* DemExportVRMLCmd::Copy() {
    Command* copy = new DemExportVRMLCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

boolean DemExportVRMLCmd::Reversible() {
    return false;
}

void DemExportVRMLCmd::Execute() {
    Style* style;
    boolean reset_caption = false;
    if (chooser_ == nil) {
	style = new Style(Session::instance()->style());
	chooser_ = new OpenFileChooser(".", WidgetKit::instance(), style);
	Resource::ref(chooser_);
	char buf[CHARBUFSIZE];
	sprintf(buf, "Export VRML to file:");
	style->attribute("caption", "");
	style->attribute("subcaption", buf);
    } else {
	style = chooser_->style();
    }
    while (chooser_->post_for(_editor->GetWindow())) {
	_editor->GetWindow()->cursor(hourglass);
	chooser_->twindow()->cursor(hourglass);
        const String* s = chooser_->selected();
	NullTerminatedString ns(*s);
	const char* name = ns.string();

	vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	    ->GetRenderWindow();
	vtkVRMLExporter* exporter = vtkVRMLExporter::New();
	exporter->SetInput(renWindow);
	exporter->SetFileName((char*)name);
	exporter->Write();
	unidraw->Update();
	break;
    }
    chooser_->unmap();
    if (reset_caption) {
	style->attribute("caption", "");
    }
}

/*****************************************************************************/

DemInitCmd::DemInitCmd(ControlInfo* info)
: Command(info) {
}

DemInitCmd::~DemInitCmd() {}

Command* DemInitCmd::Copy() {
    Command* copy = new DemInitCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void DemInitCmd::Execute() {
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
	->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkCamera* cam = vtkCamera::New();
    ren->SetActiveCamera(cam);

    vtkLight* light1 = vtkLight::New();
    light1->SetPosition(0, 1, 1);
    light1->SetFocalPoint(0, 0, 0);
    light1->SetIntensity(0.25);
    ren->AddLight(light1);
    vtkLight* light2 = vtkLight::New();
    light2->SetPosition(-1, 0, 1);
    light2->SetFocalPoint(0, 0, 0);
    light2->SetIntensity(0.25);
    ren->AddLight(light2);
#if 0
    vtkCamera* cam = ren->GetActiveCamera();
// This works but SetViewUp(0, 0, 1) doesn\'t
    cam->Elevation(-90);
#endif
}

/*****************************************************************************/

DemCameraResetCmd::DemCameraResetCmd(ControlInfo* ci)
: DemCameraCmd(ci) {
}

DemCameraResetCmd::~DemCameraResetCmd() {}

Command* DemCameraResetCmd::Copy() {
    Command* copy = new DemCameraResetCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

boolean DemCameraResetCmd::Reversible() { return false; }

void DemCameraResetCmd::Execute() {
    reset_camera();
}

/*****************************************************************************/

DemCameraUpdateCmd::DemCameraUpdateCmd(ControlInfo* ci, float px, float py,
				   float fx, float fy, boolean project)
: DemCameraCmd(ci) {
  _px = px; _py = py; _fx = fx; _fy = fy;
  _project = project;
}

DemCameraUpdateCmd::~DemCameraUpdateCmd() {}

Command* DemCameraUpdateCmd::Copy() {
    Command* copy = new DemCameraUpdateCmd(CopyControlInfo(),
					   _px, _py, _fx, _fy);
    InitCopy(copy);
    return copy;
}

boolean DemCameraUpdateCmd::Reversible() { return false; }

void DemCameraUpdateCmd::Execute() {
    update_camera(_px, _py, _fx, _fy, _project);
}

/*****************************************************************************/

DemCameraUpCmd::DemCameraUpCmd(ControlInfo* ci, float up_amount)
: DemCameraCmd(ci) {
    upamt = up_amount;
}

DemCameraUpCmd::~DemCameraUpCmd() {}

Command* DemCameraUpCmd::Copy() {
    Command* copy = new DemCameraUpCmd(CopyControlInfo(), upamt);
    InitCopy(copy);
    return copy;
}

boolean DemCameraUpCmd::Reversible() { return false; }

void DemCameraUpCmd::Execute() {
    float px, py, pz, fx, fy, fz;
    get_camera(px, py, pz, fx, fy, fz);
    set_camera(px, py, pz+upamt, fx, fy, fz+upamt);
    DemEditor* ed = (DemEditor*)GetEditor();
    ed->CameraHeight(ed->CameraHeight()+upamt);
    update_states();
}

/*****************************************************************************/

DemCameraDownCmd::DemCameraDownCmd(ControlInfo* ci, float dn_amount)
: DemCameraCmd(ci) {
    dnamt = dn_amount;
}

DemCameraDownCmd::~DemCameraDownCmd() {}

Command* DemCameraDownCmd::Copy() {
    Command* copy = new DemCameraDownCmd(CopyControlInfo(), dnamt);
    InitCopy(copy);
    return copy;
}

boolean DemCameraDownCmd::Reversible() { return false; }

void DemCameraDownCmd::Execute() {
    float px, py, pz, fx, fy, fz;
    get_camera(px, py, pz, fx, fy, fz);
    set_camera(px, py, pz-dnamt, fx, fy, fz-dnamt);
    DemEditor* ed = (DemEditor*)GetEditor();
    ed->CameraHeight(ed->CameraHeight()-dnamt);
    update_states();
}

/*****************************************************************************/

DemCameraYawLeftCmd::DemCameraYawLeftCmd(ControlInfo* ci, float yl_amount)
: DemCameraCmd(ci) {
    ylamt = yl_amount;
}

DemCameraYawLeftCmd::~DemCameraYawLeftCmd() {}

Command* DemCameraYawLeftCmd::Copy() {
    Command* copy = new DemCameraYawLeftCmd(CopyControlInfo(), ylamt);
    InitCopy(copy);
    return copy;
}

boolean DemCameraYawLeftCmd::Reversible() { return false; }

void DemCameraYawLeftCmd::Execute() {
    yaw_camera(ylamt);
}

/*****************************************************************************/

DemCameraYawRightCmd::DemCameraYawRightCmd(ControlInfo* ci, float yr_amount)
: DemCameraCmd(ci) {
    yramt = yr_amount;
}

DemCameraYawRightCmd::~DemCameraYawRightCmd() {}

Command* DemCameraYawRightCmd::Copy() {
    Command* copy = new DemCameraYawRightCmd(CopyControlInfo(), yramt);
    InitCopy(copy);
    return copy;
}

boolean DemCameraYawRightCmd::Reversible() { return false; }

void DemCameraYawRightCmd::Execute() {
    yaw_camera(-yramt);
}

/*****************************************************************************/

DemCameraDollyCmd::DemCameraDollyCmd(ControlInfo* ci, float amount)
: DemCameraCmd(ci) {
    _amt = amount;
}

Command* DemCameraDollyCmd::Copy() {
    Command* copy = new DemCameraDollyCmd(CopyControlInfo(), _amt);
    InitCopy(copy);
    return copy;
}

boolean DemCameraDollyCmd::Reversible() { return false; }

void DemCameraDollyCmd::Execute() {
    GetEditor()->GetWindow()->cursor(hourglass);
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
        ->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkCamera* camera = ren->GetActiveCamera();
    camera->Dolly(_amt);
    update_states();
    GetEditor()->GetViewer()->Draw();
}

/*****************************************************************************/

DemRepresentationCmd::DemRepresentationCmd(ControlInfo* ci, DemRep rep)
: Command(ci) {
    _rep = rep;
}

Command* DemRepresentationCmd::Copy() {
    Command* copy = new DemRepresentationCmd(CopyControlInfo(), _rep);
    InitCopy(copy);
    return copy;
}

boolean DemRepresentationCmd::Reversible() { return false; }

void DemRepresentationCmd::Execute() {
    GetEditor()->GetWindow()->cursor(hourglass);
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
        ->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkActorCollection* actors = ren->GetActors();
    vtkActor* actor;
    for (actors->InitTraversal(); actor = actors->GetNextItem(); ) {
        switch (_rep) {
	case POINTS: actor->GetProperty()->SetRepresentationToPoints(); break;
	case WIREFRAME: actor->GetProperty()->SetRepresentationToWireframe(); break;
	case SURFACE: actor->GetProperty()->SetRepresentationToSurface(); break;
	}
    }
    GetEditor()->GetViewer()->Draw();
}

/*****************************************************************************/

DemShadingCmd::DemShadingCmd(ControlInfo* ci, DemShade shade)
: Command(ci) {
    _shade = shade;
}

Command* DemShadingCmd::Copy() {
    Command* copy = new DemShadingCmd(CopyControlInfo(), _shade);
    InitCopy(copy);
    return copy;
}

boolean DemShadingCmd::Reversible() { return false; }

void DemShadingCmd::Execute() {
    GetEditor()->GetWindow()->cursor(hourglass);
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
        ->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkActorCollection* actors = ren->GetActors();
    vtkActor* actor;
    for (actors->InitTraversal(); actor = actors->GetNextItem(); ) {
        switch (_shade) {
	case FLAT: actor->GetProperty()->SetInterpolationToFlat(); break;
	case GOURAUD: actor->GetProperty()->SetInterpolationToGouraud(); break;
	case PHONG: actor->GetProperty()->SetInterpolationToPhong(); break;
	}
    }
    GetEditor()->GetViewer()->Draw();
}

/*****************************************************************************/

DemBackfaceCullCmd::DemBackfaceCullCmd(ControlInfo* ci, boolean bfc)
: Command(ci) {
    _bfc = bfc;
}

Command* DemBackfaceCullCmd::Copy() {
    Command* copy = new DemBackfaceCullCmd(CopyControlInfo(), _bfc);
    InitCopy(copy);
    return copy;
}

boolean DemBackfaceCullCmd::Reversible() { return false; }

void DemBackfaceCullCmd::Execute() {
    GetEditor()->GetWindow()->cursor(hourglass);
    vtkRenderWindow* renWindow = ((DemViewer*)GetEditor()->GetViewer())
        ->GetRenderWindow();
    vtkRendererCollection* rc = renWindow->GetRenderers();
    rc->InitTraversal();
    vtkRenderer* ren = rc->GetNextItem();
    vtkActorCollection* actors = ren->GetActors();
    vtkActor* actor;
    for (actors->InitTraversal(); actor = actors->GetNextItem(); ) {
        actor->GetProperty()->SetBackfaceCulling(_bfc);
    }
    GetEditor()->GetViewer()->Draw();
}

/*****************************************************************************/

DemStartCaptureCmd::DemStartCaptureCmd(ControlInfo* ci) : Command(ci) {
}

void DemStartCaptureCmd::Execute () {
    ((DemViewer*)GetEditor()->GetViewer())->CaptureOn();
}

/*****************************************************************************/

DemStopCaptureCmd::DemStopCaptureCmd(ControlInfo* ci) : Command(ci) {
}

void DemStopCaptureCmd::Execute () {
    ((DemViewer*)GetEditor()->GetViewer())->CaptureOff();
}

/*****************************************************************************/

DemVehicleInitCmd::DemVehicleInitCmd(ControlInfo* ci) : DemCmd(ci) {
}

void DemVehicleInitCmd::Execute () {
#ifdef HAVE_ACE
      ACE_SOCK_Stream* _socket = new ACE_SOCK_Stream;
      ACE_SOCK_Connector* connector = new ACE_SOCK_Connector;
      ACE_INET_Addr* addr = new ACE_INET_Addr(20017, "localhost");

      if (connector->connect (*_socket, *addr) == -1) {
	GetEditor()->GetWindow()->cursor(arrow);
	ACE_ERROR ((LM_ERROR, "%p\n", "open"));
	return;
      }
      ((DemEditor*)GetEditor())->VehicleSocket(_socket);
      char _com[256];
      float camera_p[3];
      float camera_f[3];
      char vsymbol[40];
      sprintf(vsymbol, "demviewer_vehicle_%c", 'a'+((DemEditor*)GetEditor())->GetCount());
      ((DemEditor*)GetEditor())->VehicleName(strdup(vsymbol));
      get_camera(camera_p[0], camera_p[1], camera_p[2],
		 camera_f[0], camera_f[1], camera_f[2]);
      sprintf(_com, "vehicle(%s);connect(%s \"localhost\" 20012);connect(%s);movevhcl(%s %f %f :abs)\n", vsymbol, vsymbol, vsymbol, vsymbol, camera_p[0],camera_p[1]);
      if (_socket->send (_com, strlen(_com)) == -1) {
	GetEditor()->GetWindow()->cursor(arrow);
	ACE_ERROR ((LM_ERROR, "%p\n", "send"));
	return;
      }
#endif /* HAVE_ACE */
}

/*****************************************************************************/

DemVehicleForwardCmd::DemVehicleForwardCmd(ControlInfo* ci) : Command(ci) {
}

void DemVehicleForwardCmd::Execute () {
#ifdef HAVE_ACE
    char* name = ((DemEditor*)GetEditor())->VehicleName();
    if (name) {
      ACE_SOCK_Stream* _socket = (ACE_SOCK_Stream*)((DemEditor*)GetEditor())->
	VehicleSocket();
      if (_socket) {
      char _com[BUFSIZ];
      sprintf(_com, "forward2d(%s 150)\n", name);
      if (_socket->send (_com, strlen(_com)) == -1) {
	GetEditor()->GetWindow()->cursor(arrow);
	ACE_ERROR ((LM_ERROR, "%p\n", "send"));
	return;
      }
      }
    }
#endif /* HAVE_ACE */
}

/*****************************************************************************/

DemVehicleBackwardCmd::DemVehicleBackwardCmd(ControlInfo* ci) : Command(ci) {
}

void DemVehicleBackwardCmd::Execute () {
#ifdef HAVE_ACE
    char* name = ((DemEditor*)GetEditor())->VehicleName();
    if (name) {
      ACE_SOCK_Stream* _socket = (ACE_SOCK_Stream*)((DemEditor*)GetEditor())->
	VehicleSocket();
      if (_socket) {
      char _com[BUFSIZ];
      sprintf(_com, "backward2d(%s 150)\n", name);
      if (_socket->send (_com, strlen(_com)) == -1) {
	GetEditor()->GetWindow()->cursor(arrow);
	ACE_ERROR ((LM_ERROR, "%p\n", "send"));
	return;
      }
      }
    }
#endif /* HAVE_ACE */
}

/*****************************************************************************/

DemVehicleTurnLeftCmd::DemVehicleTurnLeftCmd(ControlInfo* ci) : Command(ci) {
}

void DemVehicleTurnLeftCmd::Execute () {
#ifdef HAVE_ACE
    char* name = ((DemEditor*)GetEditor())->VehicleName();
    if (name) {
      ACE_SOCK_Stream* _socket = (ACE_SOCK_Stream*)((DemEditor*)GetEditor())->
	VehicleSocket();
      if (_socket) {
      char _com[BUFSIZ];
      sprintf(_com, "rotate2d(%s -15)\n", name);
      if (_socket->send (_com, strlen(_com)) == -1) {
	GetEditor()->GetWindow()->cursor(arrow);
	ACE_ERROR ((LM_ERROR, "%p\n", "send"));
	return;
      }
      }
    }
#endif /* HAVE_ACE */
}

/*****************************************************************************/

DemVehicleTurnRightCmd::DemVehicleTurnRightCmd(ControlInfo* ci) : Command(ci) {
}

void DemVehicleTurnRightCmd::Execute () {
#ifdef HAVE_ACE
    char* name = ((DemEditor*)GetEditor())->VehicleName();
    if (name) {
      ACE_SOCK_Stream* _socket = (ACE_SOCK_Stream*)((DemEditor*)GetEditor())->
	VehicleSocket();
      if (_socket) {
      char _com[BUFSIZ];
      sprintf(_com, "rotate2d(%s 15)\n", name);
      if (_socket->send (_com, strlen(_com)) == -1) {
	GetEditor()->GetWindow()->cursor(arrow);
	ACE_ERROR ((LM_ERROR, "%p\n", "send"));
	return;
      }
      }
    }
#endif /* HAVE_ACE */
}

#endif


/*****************************************************************************/

float DemVehicleGasCmd::_speed = 0.0;
float DemVehicleGasCmd::_incr = 10.0;

DemVehicleGasCmd::DemVehicleGasCmd(ControlInfo* ci) : Command(ci) {
}

void DemVehicleGasCmd::Execute () {
#ifdef HAVE_ACE
  const char* vhclname = ((DemEditor*)GetEditor())->VehicleName();
  ACE_SOCK_Stream* socket = (ACE_SOCK_Stream*)((DemEditor*)GetEditor())->
    VehicleSocket();
  if (socket) {  
    DemViewer* viewer = (DemViewer*)GetEditor()->GetViewer();
    char timestr[BUFSIZ];
    DemVehicleGasCmd::_speed += DemVehicleGasCmd::_incr;
    sprintf(timestr, "forward2d(%s %f)", vhclname, DemVehicleGasCmd::_speed);
    char com[BUFSIZ];
    sprintf(com, "timeexpr(\"%s\")\n", timestr);
    if (socket->send (com, strlen(com)) == -1) {
      GetEditor()->GetWindow()->cursor(arrow);
      ACE_ERROR ((LM_ERROR, "%p\n", "send"));
      return;
    }
  }
#endif /* HAVE_ACE */
}

/*****************************************************************************/

DemVehicleBrakeCmd::DemVehicleBrakeCmd(ControlInfo* ci) : Command(ci) {
}

void DemVehicleBrakeCmd::Execute () {
#ifdef HAVE_ACE
  const char* vhclname = ((DemEditor*)GetEditor())->VehicleName();
  ACE_SOCK_Stream* socket = (ACE_SOCK_Stream*)((DemEditor*)GetEditor())->
    VehicleSocket();
  if (socket) {  
    DemViewer* viewer = (DemViewer*)GetEditor()->GetViewer();
    char timestr[BUFSIZ];
    DemVehicleGasCmd::_speed -= DemVehicleGasCmd::_incr;
    sprintf(timestr, "forward2d(%s %f)", vhclname, DemVehicleGasCmd::_speed);
    char com[BUFSIZ];
    sprintf(com, "timeexpr(\"%s\")\n", timestr);
    if (socket->send (com, strlen(com)) == -1) {
      GetEditor()->GetWindow()->cursor(arrow);
      ACE_ERROR ((LM_ERROR, "%p\n", "send"));
      return;
    }
  }
#endif /* HAVE_ACE */
}
  
/*****************************************************************************/

DemToggleRteCaptureCmd::DemToggleRteCaptureCmd(ControlInfo* ci) : Command(ci) {
}

void DemToggleRteCaptureCmd::Execute () {
#ifdef HAVE_ACE
    const char* vhclname = ((DemEditor*)GetEditor())->VehicleName();
    const char* rtename = "demviewer_route";
    ACE_SOCK_Stream* socket = (ACE_SOCK_Stream*)((DemEditor*)GetEditor())->
	VehicleSocket();
    if (socket) {  
        DemViewer* viewer = (DemViewer*)GetEditor()->GetViewer();
        viewer->ToggleRouteCapture();
	char com[BUFSIZ];
	sprintf(com, "%s(%s %s)\n", 
		viewer->RouteCapture() ? "startroute" : "stoproute", 
		rtename, vhclname);
	if (socket->send (com, strlen(com)) == -1) {
	  GetEditor()->GetWindow()->cursor(arrow);
	  ACE_ERROR ((LM_ERROR, "%p\n", "send"));
	  return;
	}
	if (!viewer->RouteCapture()) {
	  sprintf(com, "connect(%s \"localhost\" 20010)\n", rtename);
	  if (socket->send (com, strlen(com)) == -1) {
	    GetEditor()->GetWindow()->cursor(arrow);
	    ACE_ERROR ((LM_ERROR, "%p\n", "send"));
	    return;
	  }
	}
    }
#endif /* HAVE_ACE */
}

DemShowCamStateCmd::DemShowCamStateCmd(ControlInfo* ci) : Command(ci) {}

void DemShowCamStateCmd::Execute() {
  if (GetEditor())
    ((DemEditor*)GetEditor())->ShowCameraStateWindow();
}

DemHideCamStateCmd::DemHideCamStateCmd(ControlInfo* ci) : Command(ci) {}

void DemHideCamStateCmd::Execute() {
  if (GetEditor())
    ((DemEditor*)GetEditor())->HideCameraStateWindow();
}

DemNoCullCmd::DemNoCullCmd(ControlInfo* ci) : Command(ci) {}

void DemNoCullCmd::Execute() {
  LODActor::UseFootprint(0);
  GetEditor()->GetViewer()->Draw();
}

Dem2DimFootprintCullCmd::Dem2DimFootprintCullCmd(ControlInfo* ci) : Command(ci) {}

void Dem2DimFootprintCullCmd::Execute() {
  LODActor::UseFootprint(1);
  GetEditor()->GetViewer()->Draw();
}

DemNewViewCmd::DemNewViewCmd(ControlInfo* ci) : Command(ci) {
}

Command* DemNewViewCmd::Copy() {
  DemNewViewCmd* cmd = new DemNewViewCmd();
  InitCopy(cmd);
  return cmd;
}

void DemNewViewCmd::Execute() {
  DemEditor* ed = (DemEditor*)GetEditor();
  if (ed) {
    DemKit* newKit = new DemKit();
    DemEditor* newEd = new DemEditor(ed->GetFilename(), newKit);
    unidraw->Open(newEd);
  }
}

DemMoveActorCmd::DemMoveActorCmd(Editor* ed, char* name, float x, float y,
				 float xdir, float ydir) : Command(ed)
{
  _name = name;
  _x = x;
  _y = y;
  _xdir = xdir;
  _ydir = ydir;
}

void DemMoveActorCmd::Execute() {
  DemEditor* ed = (DemEditor*)GetEditor();
  if (strcmp(ed->VehicleName(), _name)==0) {
    // camera update
    CameraCmd* cmd = new CameraCmd(_x, _y, 
				   _x + 100.0*_xdir, _y + 100.0*_ydir, false);
    cmd->SetEditor(ed);
    cmd->Execute();
  }
  else {
    // actor update

  }
}
