/*
 * Copyright (c) 1998 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <ProjUnidraw/projclasses.h>
#include <ProjUnidraw/projcmds.h>
#include <ProjUnidraw/projeditor.h>

#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapimport.h>
#include <MapUnidraw/maproute.h>
#include <MapUnidraw/mapstates.h>
#include <MapUnidraw/mapviewstate.h>
#include <MapUnidraw/mapviewer.h>

#include <OverlayUnidraw/ovarrow.h>
#include <OverlayUnidraw/ovimport.h>
#include <OverlayUnidraw/ovrect.h>

#include <Map/mapfclass.h>
#include <Map/mapfeature.h>
#include <Map/mapprim.h>

#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <UniIdraw/idarrows.h>

#include <Unidraw/catalog.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/manips.h>
#include <Unidraw/selection.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>

#include <Unidraw/Commands/edit.h>

#include <Unidraw/Graphic/damage.h>
#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Graphic/polygons.h>

#include <InterViews/cursor.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>
#include <IV-2_6/InterViews/rubline.h>
#include <IV-2_6/InterViews/rubrect.h>
#include <OS/math.h>

#include <math.h>
#include <stdio.h>
#include <strstream.h>

#ifdef HAVE_ACE
#include <ace/SOCK_Stream.h>
#include <ace/SOCK_Connector.h>
#include <ace/INET_Addr.h>
#endif

static const unit = 15;

/*****************************************************************************/

ClassId ProjShowCmd::GetClassId () { return PROJSHOW_CMD; }

boolean ProjShowCmd::IsA (ClassId id) {
    return PROJSHOW_CMD == id || Command::IsA(id);
}

ProjShowCmd::ProjShowCmd (ControlInfo* c) : Command(c) { }
ProjShowCmd::ProjShowCmd (Editor* ed) : Command(ed) { }

Command* ProjShowCmd::Copy () {
    Command* copy = new ProjShowCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void ProjShowCmd::Execute () {
#if 0
    ProjEditor* editor = (ProjEditor*)GetEditor();
    editor->GetWindow()->cursor(hourglass);
    GraphicComps* top = (GraphicComps*)GetEditor()->GetComponent();
    Clipboard* clb = new Clipboard();
    
    MapViewState* viewstate = (MapViewState*)editor->GetMapViewState();
    MapFeatureClass* mapfclass = new MapFeatureClass(nil, "utmfeature");
    if (!mapfclass) return;

    int nfeats = mapfclass->nfeats();
    for (int i = 0; i<nfeats; i++) {
	MapFeature* feature = mapfclass->feature(i);
	if (feature->face_primitive()->npts()<0) continue;
	if (!viewstate->themegs(feature)) continue;
	MapFeatureComp* mfc = new MapFeatureComp(feature);
	if (mfc) 
	    clb->Append(mfc);
    }
    
    PasteCmd* pastecmd = new PasteCmd(editor, clb);
    pastecmd->Execute();
    Selection* s = GetEditor()->GetViewer()->GetSelection();
    s->Clear();
#endif
}

boolean ProjShowCmd::Reversible () { return true; }


/*****************************************************************************/

#ifdef HAVE_ACE

ClassId ProjImportByServCmd::GetClassId () { return PROJIMPORTBYSERV_CMD; }

boolean ProjImportByServCmd::IsA (ClassId id) {
    return PROJIMPORTBYSERV_CMD == id || Command::IsA(id);
}

ProjImportByServCmd::ProjImportByServCmd (ControlInfo* c) : Command(c) { 
  _impcmd = nil;
}
ProjImportByServCmd::ProjImportByServCmd (Editor* ed) : Command(ed) { 
  _impcmd = nil;
}

Command* ProjImportByServCmd::Copy () {
    Command* copy = new ProjImportByServCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void ProjImportByServCmd::ServHost (const char* host) {
  _host = (char*)host;
}

void ProjImportByServCmd::ServPort (unsigned short port) {
  _port = port;
}

void ProjImportByServCmd::ServCom (char* com) {
  _com = com;
}

void ProjImportByServCmd::ResultImportCmd (OvImportCmd* impcmd) {
  _impcmd = impcmd;
}

void ProjImportByServCmd::Execute () {
  GetEditor()->GetWindow()->cursor(hourglass);

  ACE_SOCK_Stream* _socket = new ACE_SOCK_Stream;
  ACE_SOCK_Connector* connector = new ACE_SOCK_Connector;
  ACE_INET_Addr* addr = new ACE_INET_Addr(_port, _host);

  if (connector->connect (*_socket, *addr) == -1) {
    GetEditor()->GetWindow()->cursor(arrow);
    ACE_ERROR ((LM_ERROR, "%p\n", "open"));
    return;
  }
  if (_socket->send (_com, strlen(_com)) == -1) {
    GetEditor()->GetWindow()->cursor(arrow);
    ACE_ERROR ((LM_ERROR, "%p\n", "send"));
    return;
  }
  if (_impcmd) {
    filebuf* _filebuf = new filebuf();
    boolean valid = _filebuf->attach(_socket->get_handle()) != 0;
    istream*_inptr = new istream(_filebuf);
    ((OverlayIdrawComp*)GetEditor()->GetComponent())->ResetIndexedGS();
    _impcmd->instream(_inptr);
    _impcmd->Execute();
  }
  if (_socket->send ("quit\n", strlen("quit\n")) == -1) {
    GetEditor()->GetWindow()->cursor(arrow);
    ACE_ERROR ((LM_ERROR, "%p\n", "send"));
    return;
  }
}

boolean ProjImportByServCmd::Reversible () { return false; }

/****************************************************************************/

ProjMapImportByServCmd::ProjMapImportByServCmd (ControlInfo* c) : ProjImportByServCmd(c) { }
ProjMapImportByServCmd::ProjMapImportByServCmd (Editor* ed) : ProjImportByServCmd(ed) { }

Command* ProjMapImportByServCmd::Copy () {
    Command* copy = new ProjMapImportByServCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void ProjMapImportByServCmd::Execute () {
  Catalog* catalog = unidraw->GetCatalog();
  const char* portstr = catalog->GetAttribute("mapservport");
  u_short MAPSERV_PORT = atoi(portstr);
  const char *const MAPSERV_HOST = catalog->GetAttribute("mapservhost");
  const char *mapserv_host = MAPSERV_HOST;
  u_short mapserv_port  = MAPSERV_PORT;
  ServHost(mapserv_host);
  ServPort(mapserv_port);
  ServCom("print(loadfc(\"/data/maps/vpf/killeen\" \"us064462\" \"trn\" \"trdline\") :paint)\n");
  ProjEditor* editor = (ProjEditor*)GetEditor();
  MapImportCmd* _import_cmd = new MapImportCmd(editor);
  ResultImportCmd(_import_cmd);
  ProjImportByServCmd::Execute();
}

#endif

/****************************************************************************/

ClassId ProjPosTool::GetClassId () { return PROJPOS_TOOL; }

boolean ProjPosTool::IsA (ClassId id) {
    return PROJPOS_TOOL == id || GraphicCompTool::IsA(id);
}

Tool* ProjPosTool::Copy () { return new ProjPosTool(CopyControlInfo()); }

ProjPosTool::ProjPosTool(ControlInfo* info)
:GraphicCompTool()
{
    if (info)
	SetControlInfo(info);
    ArrowLine* ml = new ArrowLine(
        0, 0, unit, unit, true, false, 2., stdgraphic
    );
    Catalog* catalog = unidraw->GetCatalog();
    PSBrush* br = catalog->FindBrush(0xffff, 2);
    ml->SetBrush(br);
    PSColor* col = catalog->FindColor("Red");
    ml->SetColors(col, col);
    ArrowLineOvComp* mlComp = new ArrowLineOvComp(ml);
    Init(mlComp);
}

Command* ProjPosTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    MapEditor* ed = (MapEditor*) dm->GetViewer()->GetEditor();
    Transformer* rel = dm->GetTransformer();
    MacroCmd* cmd = new MacroCmd(ed);
    Transformer* trans = ed->GetMapViewState()->deftrans();

    RubberLine* rl = (RubberLine*) dm->GetRubberband();

    Coord px0, py0, px1, py1;
    int pt;
    rl->GetCurrent(px0, py0, px1, py1);

    if (rel != nil) {
      rel = new Transformer(rel);
      rel->Invert();
    }

    if (px0 != px1 || py0 != py1) {
        float fpx, fpy, ffx, ffy;
	rel->transform(px0, py0, fpx, fpy);
	rel->transform(px1, py1, ffx, ffy);
	printf("rl: %d %d %d %d; trl: %f %f %f %f\n",
	       px0, py0, px1, py1, fpx, fpy, ffx, ffy);
	if (trans) {
	  float tpx, tpy, tfx, tfy;
	  trans->inverse_transform(fpx, fpy, tpx, tpy);
	  trans->inverse_transform(ffx, ffy, tfx, tfy);
	  printf("UTM: P1: %fE %fN; P2: %fE %fN\n", tpx, tpy, tfx, tfy);
	}

	if (ed->GetCameraComp()) {
	  Viewer* viewer = ed->GetViewer();    
	  OverlayView* view =
	    ((OverlayComp*)ed->GetCameraComp())->FindView(ed->GetViewer());
	  ArrowLineOvComp* comp = (ArrowLineOvComp*)ed->GetCameraComp();
	  ArrowLine* aline = comp->GetArrowLine();
	  aline->SetTransformer(rel);
	  //	      aline->SetOriginal(px0, py0, px1, py1);

	  ArrowLine* alinev = (ArrowLine*)(view->GetGraphic());
	  viewer->GetDamage()->Incur(alinev);
	  alinev->SetTransformer(rel);
	  alinev->SetOriginal(px0, py0, px1, py1);
	  viewer->GetDamage()->Incur(alinev);
	  unidraw->Update();
	}
	else {
	  Catalog* catalog = unidraw->GetCatalog();
	  ArrowLine* aline = new ArrowLine(px0, py0, px1, py1, false, true,
					   1.25, stdgraphic);
	  PSBrush* br = catalog->FindBrush(0xffff, 2);
	  aline->SetBrush(br);
	  PSColor* col = catalog->FindColor("Red");
	  aline->SetColors(col, col);
	  aline->SetTransformer(rel);
	  ArrowLineOvComp* comp = new ArrowLineOvComp(aline);
	  Command* pcmd = new PasteCmd(ed, new Clipboard(comp));
	  pcmd->Execute();
	  ed->GetSelection()->Clear();
	  ed->SetCameraComp(comp);
	}
    }
    return cmd;
}
